/*----------------------------------------------------------------------
  File    : Combined1.java
  Contents: split a (sub-)pointgon with triangles and paths
  Author  : Christian Borgelt
  History : 2006.01.27 file created from file Path.java
            2006.02.05 made a subclass of Path rather than Splitter
            2006.02.06 first version of function split finished
            2006.02.07 missing intersection check added to method valid
----------------------------------------------------------------------*/
package pointgon;

/*--------------------------------------------------------------------*/
/** Class for first combined path splitting.
 *  @since  2006.01.27 (Christian Borgelt) */
/*--------------------------------------------------------------------*/
public class Combined1 extends Path {

  /*------------------------------------------------------------------*/
  /*  constants                                                       */
  /*------------------------------------------------------------------*/
  private static final long serialVersionUID = 0x00010000;

  /*------------------------------------------------------------------*/
  /*  instance variables                                              */
  /*------------------------------------------------------------------*/
  /** flag for a one-path subproblem */
  private boolean onep;

  /*------------------------------------------------------------------*/
  /** Check added path edge(s).
   *  @return whether the added path edge(s) is/are valid
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private boolean valid ()
  {                             /* --- check added path edge(s) */
    int    i, k, e;             /* loop variables */
    Vertex v, p, q, x, y;       /* vertices of edge/triangle, buffers */

    if (this.plen > 2) {        /* if there is more than one hole */
      p = this.path[this.plen-3];  /* get and check the edge that */
      q = this.path[this.plen-2];  /* was added to the sequence */
      return (this.wgts[p.id][q.id] >= 0)
          && !this.isects(p, q);
    }                           /* return whether edge is valid */
    p = this.key[0];            /* get the anchor and */
    v = this.path[0];           /* the first vertex of the path */
    if (!this.onep)             /* if this is a two-path subproblem */
      return (this.wgts[p.id][v.id] >= 0)
          &&  this.inside(p,v)  /* return whether edge is valid */
          && !this.isects(p,v); /* (leads inside and no intersection) */
    q = this.key[i = k = 1];    /* get the base of the triangle */
    if (q == null) { q = p; p = this.key[i = this.cnt-1]; k = -1; }
    if (!this.valid(p, q, v))   /* if the triangle is invalid, */
      return false;             /* abort with failure */
    x = this.key[i];            /* get the right base vertex */
    if (v.compareTo(x) <= 0)    /* if left or middle triangle, */
      return true;              /* there can be no path intersection */
    y = this.key[i += k];       /* get key vertex right of the base */
    if ((y == v) || (y == null))/* if it is the tip or one edge key, */
      return true;              /* there can be no path intersection */
    if ((this.trtab == null)    /* if there is no precomputed info. */
    &&  y.isInside(p, q, v))    /* and the next key vertex is inside, */
      return false;             /* the triangle is invalid */
    if (this.key[0].isectsX(v, x, y))
      return false;             /* check left side against key edge */
    while (true) {              /* traverse the remaining key edges */
      x = y; y = this.key[i += k]; /* get vertices of next edge */
      if (y == null) return true;  /* and check for the key end */
      if (y == v)    break;     /* check for the triangle tip */
      if (p.isectsX(v, x, y) || q.isectsX(v, x, y))
        return false;           /* check intersection of the key edge */
    }                           /* with either side of the triangle */
    return true;                /* return 'path is valid' */
  }  /* valid() */

  /*------------------------------------------------------------------*/
  /** Compute the number of vertices between two given vertices.
   *  @param  p   the first  vertex
   *  @param  q   the second vertex
   *  @return the number of vertices between the given vertices
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private int vertcnt (Vertex p, Vertex q)
  {                             /* --- compute number of vertices */
    int n;                      /* number of vertices between p and q */
    if (this.dir > 0) n = q.id -p.id;
    else              n = p.id -q.id;
    if (n <= 0) n += this.verts.length;
    return n-1;                 /* return the number of vertices */
  }  /* vertcnt() */            /* between the vertices p and q */

  /*------------------------------------------------------------------*/
  /** Create a subproblem splitter.
   *  @param  mwt   the minimum weight traingulation object
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected Combined1 (MWT mwt)
  { super(mwt); this.triangle = new Vertex[3]; }

  /*------------------------------------------------------------------*/
  /** Initialize a subproblem splitter.
   *  @param  key      the key  of the subproblem
   *  @param  cnt      the size of the subproblem
   *  @param  holes    the holes of the subproblem
   *  @param  holecnt  the number of holes of the subproblem
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void init (Vertex[] key,   int cnt,
                       Vertex[] holes, int holecnt)
  {                             /* --- init. a subproblem splitter */
    int    n;                   /* number of perimeter vertices */
    Vertex v, p, q;             /* subproblem anchor, adjacent verts. */

    this.key     = key;         /* note the key */
    this.cnt     = cnt;         /* of the subproblem */
    this.holes   = holes;       /* note the holes that are (possibly) */
    this.holecnt = holecnt;     /* inside this subproblem */
    this.pdir    = 0;           /* init. the split state */
    this.state   = 0;           /* (no splits created yet) */
    this.bestlen = 0;           /* init. the best path length */
    n = this.verts.length;      /* get modulo for id computations */
    v = key[0];                 /* get the subproblem anchor */
    p = key[1];                 /* and the key elements that are */
    q = key[cnt-1];             /* adjacent to anchor (2nd and last) */
    this.frst = (p != null) ? p : this.verts[(v.id +n +this.dir) % n];
    this.last = (q != null) ? q : this.verts[(v.id +n -this.dir) % n];
    this.onep = ((p == null) || (q == null));
    if (this.onep) { this.idxs[0] = -1; this.state = 1; }
    if ((holecnt <= 0) || !this.onep) this.lftpos = this.rgtpos = 0;
    else this.getpos(key[0], (q != null) ? q : p);
  }  /* init() */               /* get hole positions for paths */

  /*------------------------------------------------------------------*/
  /** Initialize a subproblem splitter.
   *  @param  key      the key  of the subproblem
   *  @param  cnt      the size of the subproblem
   *  @param  spobj    the splitter object
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void init (Vertex[] key, int cnt, Object spobj)
  {                             /* --- init. a subproblem splitter */
    this.onep = ((key[1] == null) || (key[cnt-1] == null));
    super.init(key,cnt,spobj);  /* use superclass initialization */
  }  /* init() */

  /*------------------------------------------------------------------*/
  /** Create next subproblem split.
   *  @return whether a new split was created
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected boolean next ()
  {                             /* --- create next subproblem split */
    int     i;                  /* index of hole vertex */
    Vertex  v, p, q;            /* key/path/perimeter vertices */
    boolean valid;              /* whether prec. sequence was valid */

    /* --- path that cuts off the anchor --- */
    if (this.state == 0) {      /* if no path has been created yet */
      this.plen  = 0;           /* create path cutting off anchor and */
      this.state = 1;           /* switch to paths starting at anchor */
      p = this.frst;            /* get the two vertices that */
      q = this.last;            /* are adjacent to the anchor */
      if ((this.wgts[p.id][q.id] >= 0)
      &&   this.validcut(p, q)  /* if the only path edge is valid, */
      &&  !this.isects(p, q))   /* the cut is valid (empty triangle), */
        return true;            /* and edge does not intersect paths, */
    }                           /* return 'next split created' */

    if (this.pdir == 0) {       /* if no holes processed yet */
      this.plen = 1;            /* path contains only the end vertex */

      /* --- build triangles over first edge --- */
      if (this.onep) {          /* if this is a one-path subproblem */
        p = this.key[0];        /* get the two key vertices */
        q = this.key[1];        /* that form the triangle base */
        if (q == null) { q = p; p = this.key[this.cnt-1]; }
        while (true) {          /* triangle tip search loop */
          if (this.idxs[0] < 0){/* if in the perimeter vertices */
            v = this.nextEnd(); /* get the next perimeter vertex */
            if (v == null) {    /* if none, switch to the holes */
              this.idxs[0] = this.rgtpos; continue; }
            if (this.valid())   /* if the triangle is valid, */
              return true; }    /* return 'next split created' */
          else {                /* if in the hole vertices */
            i = --this.idxs[0]; /* get the index of the next hole */
            if (i < this.lftpos)/* if all holes in key range have */
              break;            /* been processed, abort the loop */
            this.path[0] = v = this.holes[i];
            if (this.valid(p, q, v))
              return true;      /* get the next hole vertex as tip */
          }                     /* and if the triangle is valid, */
        } }                     /* return 'next split created' */

      /* --- single edge path starting at the anchor --- */
      else {                    /* if this is a two-path subproblem */
        p = this.key[0];        /* get the subproblem anchor */
        while (true) {          /* end vertex search loop */
          q = this.nextEnd();   /* get the next end of the path */
          if (q == null) break; /* if no more end vertices, abort */
          if ((q != this.frst)  /* edges to vertices adjacent */
          &&  (q != this.last)  /* to the anchor are not allowed */
          &&  (this.wgts[p.id][q.id] >= 0)
          &&   this.inside(p,q) /* if the edge of the path is valid, */
          &&  !this.isects(p,q))/* leads to the inside of pointgon, */
            return true;        /* and does not intersect a path, */
        }                       /* and return 'next split created' */
      }                         /* (single edge path from anchor) */

      if (this.holecnt <= 0)    /* if there are no holes, */
        return false;           /* all splits have been generated */
      do {                      /* hole sequence generation loop */
        if (!this.nextHoles(false))
          return false;         /* check for another hole sequence */
      } while (!this.valid());  /* while hole sequence is invalid */
      this.state = 1;           /* reinitialize the path state */
    }                           /* for the end vertex traversal */

    /* --- multi-edge path starting at the anchor --- */
    while (true) {              /* path generation loop */
      q = this.nextEnd();       /* get the next end vertex */
      if (q != null) {          /* if there is another end vertex */
        p = this.path[this.plen-2];  /* get preceding path vertex */
        if ((this.wgts[p.id][q.id] < 0)
        ||  (q.compareTo(p) != this.pdir)
        ||  this.isects(p, q))  /* if the edge is invalid, has wrong */
          continue;             /* dir, or intersects paths, skip it */
        return true;            /* return 'next split created' */
      }                         /* (multi-edge path from anchor) */
      valid = true;             /* preceding hole sequence was valid */
      do {                      /* hole sequence generation loop */
        if (!this.nextHoles(valid))
          return false;         /* check for another hole sequence */
        valid = this.valid();   /* check validity of hole sequence */
      } while (!valid);         /* while hole sequence is invalid */
      this.state = 1;           /* reinitialize the path state */
    }                           /* for the end vertex traversal */
  }  /* next() */

  /*------------------------------------------------------------------*/
  /** Split a subproblem with the current split.
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void split ()
  {                             /* --- split subproblem with path */
    int     i, k, n, d, e;      /* loop variables, buffers */
    Vertex  v, p, q, x, z;      /* buffers for vertices */

    /* --- path cuts off the anchor --- */
    if (this.plen <= 0) {       /* if the path cuts off the anchor */
      this.lftkey[0] = this.key[0];    /* the left subproblem */
      this.lftkey[1] = p = this.frst;  /* is the cut-off triangle */
      this.lftkey[2] = q = this.last;
      this.lftcnt = this.lftall = 3;
      k = 0;                    /* default: no perimeter section */
      this.rgtcnt = this.rgtall = this.cnt-1;
      if (p.compareTo(q) < 0) { /* if the path is left to right */
        for (i = this.cnt; --i > 0; ) {
          this.rgtkey[i-1] = v = this.key[i];
          if (v == null) k = i; /* copy the remaining key vertices */
        } }                     /* and find a perimeter section */
      else {                    /* if the path is right to left */
        this.rgtkey[0] = q;     /* make last key vertex new anchor */
        for (i = this.cnt; --i > 0; ) {
          this.rgtkey[i] = v = this.key[i];
          if (v == null) k = i; /* copy the remaining key vertices */
        }                       /* and find a perimeter section */
      }                         /* (for the number of vertices) */
      if (k > 0)                /* add vertices in perimeter section */
        this.rgtall += this.vertcnt(this.key[k-1], this.key[k+1]) -1;
      return;                   /* subproblems have been built, */
    }                           /* so abort the function */

    /* --- path starts at the anchor --- */
    if (!this.onep) {           /* if this is a two-path subproblem */
      this.rgtkey[0] = this.key[0];  /* always copy subproblem anchor */
      z = this.path[this.plen-1];    /* get the end of the path and */
      d = e = 0;                /* init. span and path end position */
      n = this.verts.length;    /* get modulo for id computations */
      for (i = 0; true; ) {     /* traverse the key elements */
        v = this.key[++i];      /* copy all vertices in the key up to */
        if (v == z) break;      /* but excluding the end of the path */
        if (v != null) {        /* if not at the perimeter section */
          if ((v.id < n) && (e <= 0))
            e = 1; }            /* note the end of the ccw. path */
        else {                  /* if at the perimeter section */
          p = this.key[(i -this.dir) % this.cnt];
          e = this.key[(i +this.dir) % this.cnt].id -p.id;
          if (e <= 0) e += n;   /* (e is number of edges in section) */
          if (z.id < n) {       /* if path ends at perimeter vertex */
            d = (z.id +n -p.id) % n;
            if (this.dir < 0) d = e-d;
            if (d < e) break;   /* if end of path in section, abort */
          }                     /* (d states position in section) */
        }
        this.rgtkey[i] = v;     /* copy the current key vertex */
      }                         /* (partially build right subproblem) */
      this.rgtcnt = i;          /* set the number of copied vertices */
      if (v != null)            /* if end not in perimeter section, */
        this.rgtall = (e > 2) ? e-2 : 0;    /* get the section span */
      else if (d <= 1)          /* if no vertex left in section, */
        this.rgtall = 0;        /* there are no additional vertices */
      else {                    /* add perimeter section to key */
        this.rgtall = d-2; this.rgtkey[this.rgtcnt++] = null; }
      for (k = this.plen; --k >= 0; )
        this.rgtkey[this.rgtcnt++] = this.path[k];
      this.rgtall += this.rgtcnt;   /* add the reversed path */
      this.lftall = 0;          /* initialize vertex counter */
      this.lftkey[0] = this.key[0];
      for (k = this.plen; --k >= 0; )
        this.lftkey[k+1] = this.path[k];
      this.lftcnt = this.plen+1;/* start with the path (forward) */
      if ((v != null)           /* if end not in perimeter section */
      ||  (d+1 >= e))           /* or no vertex left in section, */
        e = 0;                  /* there are no additional vertices */
      else {                    /* add perimeter section to key */
        e -= d+2; this.lftkey[this.lftcnt++] = null; }
      while (++i < this.cnt) {  /* add the rest of the current key */
        this.lftkey[this.lftcnt++] = v = this.key[i];
        if (v != null) continue;/* if not at perimeter section, skip */
        p = this.key[(i -this.dir) % this.cnt];
        e = this.key[(i +this.dir) % this.cnt].id -p.id;
        if (e <= 0) e += n;     /* (e: number of edges in section) */
        e -= 2;                 /* subtract vertices that are */
      }                         /* already on the subproblem key */
      this.lftall = this.lftcnt +e;
      return;                   /* subproblems have been built, */
    }                           /* so abort the function */

    /* --- left triangle with/without path --- */
    z = this.path[this.plen-1]; /* get the end of the path */
    v = this.path[0];           /* get the triangle tip/path start */
    p = this.key[0];            /* and the two vertices that form */
    q = this.key[1];            /* the base of the triangle, */
    if (q != null) { x = p; }   /* also get vertex after perim. sect. */
    else           { x = this.key[2]; q = this.key[this.cnt-1]; }
    if (v.compareTo(p) < 0) {   /* if path is left of the anchor */
      this.lftkey[k = 0] = z;   /* end of the path is new anchor */
      this.lftall = this.vertcnt(z, x);
      if (this.lftall > 0) {  /* add perimeter section if necessary */
        this.lftkey[++k] = null; this.lftall--; }
      this.triangle[0] = v;     /* the tip is the triangle's anchor */
      if (this.key[1] == null){ /* if the path is on the cw side */
        this.triangle[1] = q;   /* the triangle points downwards */
        this.triangle[2] = p;   /* (build counterclockwise path) */
        System.arraycopy(this.key,  2, this.lftkey, ++k, this.cnt-2);
        k += this.cnt-2;        /* add the forward subproblem key */
        System.arraycopy(this.path, 0, this.lftkey,   k, this.plen-1);
        k += this.plen-1;       /* add the forward path (decreasing) */
        this.lftall += this.lftcnt = k;  /* set the subproblem sizes */
        for (k = 0, i = this.plen; --i >= 0; )
          this.rgtkey[k++] = this.path[i];
        this.rgtkey[k++] = p; } /* right subproblem is reversed path */
      else {                    /* if the path is on the ccw side */
        this.triangle[1] = p;   /* the triangle points upwards */
        this.triangle[2] = q;   /* (build counterclockwise path) */
        this.lftkey[++k] = p;   /* add the old subproblem anchor */
        System.arraycopy(this.path, 0, this.lftkey, ++k, this.plen-1);
        k += this.plen-1;       /* add the forward path (decreasing) */
        this.lftall += this.lftcnt = k;  /* set the subproblem sizes */
        for (k = 0, i = this.plen; --i >= 0; )  /* start with the */
          this.rgtkey[k++] = this.path[i];      /* reversed path */
        System.arraycopy(this.key, 1, this.rgtkey, k, this.cnt-2);
        k += i = this.cnt-2;    /* add the forward subproblem key */
        p = this.key[i];        /* get the end of the current key */
      }                         /* for calling the vertex counter */
      this.rgtall = this.vertcnt(p, z);
      if (this.rgtall > 0) {    /* add perimeter section if necessary */
        this.rgtkey[k++] = null; this.rgtall--; }
      this.rgtall += this.rgtcnt = k; /* set the subproblem sizes */
      if (this.rgtall < 3) this.rgtall = this.rgtcnt = 0;
      if (this.lftall < 3) this.lftall = this.lftcnt = 0;
      return;                   /* check for proper subproblems */
    }                           /* and abort the function */

    /* --- right triangle with/without path --- */
    if (v.compareTo(q) > 0) {   /* if path is right of triangle base */
      this.triangle[0] = p;     /* old anchor is triangle's anchor */
      if (this.key[1] == null){ /* if the path is on the cw side */
        this.triangle[1] = v;   /* the triangle points downwards */
        this.triangle[2] = q;   /* (build counterclockwise walk) */
        this.lftkey[0]   = q;   /* last vertex is the left anchor */
        System.arraycopy(this.path, 0, this.lftkey, 1, this.plen);
        k = this.plen+1;        /* add the forward path (increasing) */
        x = this.key[2];        /* get the start of the cw side path */
        if ((z.id < this.verts.length)
        &&  (z != x)) {         /* if path end is in perim. section */
          this.lftall = this.vertcnt(z, x);
          if (this.lftall > 0){ /* add perimeter section if necessary */
            this.lftkey[k++] = null; this.lftall--; }
          x = z;                /* replace the end of the key and */
          n = 1; }              /* get the key index for copying */
        else {                  /* if path end lies in the key */
          this.lftall = 0;      /* there cannot be a perim. section */
          for (n = this.cnt-1; this.key[--n] != z; );
        }                       /* find the path end in the key */
        System.arraycopy(this.key, n+1, this.lftkey, k, this.cnt-n-2);
        k += this.cnt-n-2;      /* copy (part of) the subproblem key */
        this.lftall += this.lftcnt = k;  /* set the subproblem sizes */
        this.rgtkey[k = 0] = p; /* right subproblem has old anchor */
        this.rgtall = this.vertcnt(p, x);
        if (this.rgtall > 0) {  /* add perimeter section if necessary */
          this.rgtkey[++k] = null; this.rgtall--; }
        for (i = 1; ++i < n; )  /* add other part of subproblem key */
          this.rgtkey[++k] = this.key[i];
        for (i = this.plen; --i >= 0; )     /* add the */
          this.rgtkey[++k] = this.path[i];  /* reversed path */
        this.rgtall += this.rgtcnt = ++k; }
      else {                    /* if the path is on the ccw side */
        this.triangle[1] = q;   /* the triangle points upwards */
        this.triangle[2] = v;   /* (build counterclockwise walk) */
        x = this.key[this.cnt-2];  /* get end of the ccw side path */
        if ((z.id < this.verts.length)
        &&  (z != x)) {         /* if path end is in perim. section */
          System.arraycopy(this.key, 1, this.rgtkey, 0, this.cnt-2);
          k = this.cnt-2;       /* start with the subproblem key */
          this.rgtall = this.vertcnt(x, z);
          if (this.rgtall > 0){ /* add perimeter section if necessary */
            this.rgtkey[k++] = null; this.rgtall--; }
          x = z;                /* replace the end of the key and */
          n = this.cnt-1; }     /* get the key index for copying */
        else {                  /* if path end lies in the key */
          this.rgtall = k = 0;  /* there is no perimeter section */
          for (v = this.key[1]; v != z; v = this.key[++k+1])
            this.rgtkey[k] = v; /* copy (part of) the subprob. key */
          n = k+2;              /* get key index for copying */
        }                       /* (n is key index of path end) */
        for (i = this.plen; --i >= 0; )     /* add the */
          this.rgtkey[k++] = this.path[i];  /* reversed path */
        this.rgtall += this.rgtcnt = k;     /* set subproblem sizes */
        this.lftkey[0] = p;     /* left subproblem has old anchor */
        System.arraycopy(this.path, 0, this.lftkey, 1, this.plen);
        k = this.plen+1;        /* add the forward path (increasing) */
        while (n < this.cnt-1)  /* add other path of subproblem key */
          this.lftkey[k++] = this.key[n++];
        this.lftall = this.vertcnt(x, p);
        if (this.lftall > 0) {  /* add perimeter section if necessary */
          this.lftkey[k++] = null; this.lftall--; }
        this.lftall += this.lftcnt = k;
      }                         /* set subproblem sizes */
      if (this.rgtall < 3) this.rgtall = this.rgtcnt = 0;
      if (this.lftall < 3) this.lftall = this.lftcnt = 0;
      return;                   /* check for proper subproblems */
    }                           /* and abort the function */

    /* --- middle triangle (tip is a hole) --- */
    if (v.id >= this.verts.length){ /* if the tip is a hole vertex */
      this.lftkey[0] = p;       /* add triangle tip next to anchor */
      if (this.key[1] == null){ /* if the path is on the cw side */
        this.lftall = this.vertcnt(p, this.key[2]);
        this.lftkey[this.cnt] = v; i = 1; }
      else {                    /* if the path is on the ccw side */
        this.lftall = this.vertcnt(this.key[this.cnt-2], p);
        this.lftkey[1]        = v; i = 2;
      }                         /* get form specific parameters */
      System.arraycopy(this.key, 1, this.lftkey, i, this.cnt-1);
      this.lftcnt = k = this.cnt+1;
      this.lftall = (this.lftall > 0) ? this.lftall +k-1 : k;
      this.triangle[0] = p;     /* copy the remaining subproblem key */
      this.triangle[i] = v;     /* and build the cut-out triangle */
      this.triangle[(i & 1) +1] = q;
      this.rgtcnt = this.rgtall = 0;
      return;                   /* set the right subproblem empty */
    }                           /* and abort the function */

    /* --- middle triangle (tip is on perimeter) --- */
    this.triangle[0] = p;       /* subp. anchor is triangle anchor */
    if (this.key[1] == null) {  /* if the path is on the cw side */
      this.triangle[1] = v;     /* the triangle points downwards */
      this.triangle[2] = q;     /* (build counterclockwise walk) */
      this.lftkey[k = 0] = v;   /* the left subp. starts at the tip */
      this.lftall = this.vertcnt(v, this.key[2]);
      if (this.lftall > 0) {    /* add perimeter section if necessary */
        this.lftkey[++k] = null; this.lftall--; }
      System.arraycopy(this.key, 2, this.lftkey, ++k, this.cnt-2);
      k += this.cnt-2;          /* copy the forward subproblem key */
      this.lftall += this.lftcnt = k;    /* set subproblem sizes */
      this.rgtall = this.vertcnt(p, v);
      if (this.rgtall <= 0)     /* if right subproblem is empty */
        this.rgtcnt = this.rgtall = 0;
      else {                    /* if right subproblem is not empty */
        this.rgtkey[0] = p;     /* its anchor is the old anchor */
        this.rgtkey[1] = null;  /* it contains only one edge and */
        this.rgtkey[2] = v;     /* a consecutive perimeter piece */
        this.rgtcnt = 3; this.rgtall += 2;
      } }                       /* set the subproblem sizes */
    else {                      /* if the path is on the ccw side */
      this.triangle[1] = v;     /* the triangle points downwards */
      this.triangle[2] = q;     /* (build counterclockwise walk) */
      this.lftall = this.vertcnt(v, p);
      if (this.lftall <= 0)     /* if left subproblem is empty */
        this.lftcnt = this.lftall = 0;
      else {                    /* if left subproblem is not empty */
        this.lftkey[0] = p;     /* its anchor is the old anchor */
        this.lftkey[1] = v;     /* it contains only one edge and */
        this.lftkey[2] = null;  /* a consecutive perimeter piece */
        this.lftcnt = 3; this.lftall += 2;
      }                         /* set the subproblem sizes */
      this.rgtkey[0] = v;       /* the right subp. starts at the tip */
      System.arraycopy(this.key, 1, this.rgtkey, 1, this.cnt-2);
      this.rgtall = this.vertcnt(this.key[this.cnt-2], v);
      k = this.cnt-1;           /* copy the forward subproblem key */
      if (this.rgtall > 0) {    /* add perimeter section if necessary */
        this.rgtkey[k++] = null; this.rgtall--; }
      this.rgtall += this.rgtcnt = k;
    }                           /* set the subproblem sizes */
    if (this.rgtall < 3) this.rgtall = this.rgtcnt = 0;
    if (this.lftall < 3) this.lftall = this.lftcnt = 0;
  }  /* split() */              /* check for proper subproblems */

  /*------------------------------------------------------------------*/
  /** Get the current triangle.
   *  @return the current triangle
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected Vertex[] triangle ()
  { return (this.onep) ? this.triangle : null; }

  /*------------------------------------------------------------------*/
  /** Compute the weight of the current split.
   *  @return the weight of the current split
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected double weight ()
  {                             /* --- compute weight of split */
    int    i, k, n;             /* loop variables, number of vertices */
    Vertex v, p, q;             /* corners of a cut-off triangle */
    double wgt = 0;             /* weight of the path/triangle */

    if (this.plen == 0)         /* if path cuts off the anchor */
      return -this.wgts[this.frst.id][this.last.id];
    for (i = this.plen; --i > 0; )  /* sum edge lengths of the path */
      wgt -= this.wgts[this.path[i].id][this.path[i-1].id];
    p = this.key[0];            /* get the anchor of the subproblem */
    v = this.path[0];           /* and the triangle tip/path start */
    if (!this.onep)             /* if this is a two-path subproblem */
      return wgt -this.wgts[p.id][v.id];
    q = this.key[i = k = 1];    /* get the base of the triangle */
    if (q == null) { q = this.key[i = this.cnt-1]; k = -1; }
    wgt += this.wgts[p.id][q.id];  /* get weight of triangle base */
    n = this.verts.length;      /* get modulo for id computations */
    if (v.id == (p.id +n -k) % n)        /* if triangle cuts off */
      return wgt +this.wgts[p.id][v.id];           /* the anchor */
    if ((v == this.key[i+k])    /* if tip is vertex after base */
    ||  ((q.id < n)             /* or there is only one key edge */
    &&   (v.id == (q.id +n +k) % n)))   /* and triangle cuts off */
      return wgt +this.wgts[q.id][v.id];          /* the key end */
    return wgt;                 /* add weight of opp. triangle side */
  }  /* weight() */             /* return the computed split weight */

  /*------------------------------------------------------------------*/

  protected Edge[] edges ()
  {                             /* --- get edges of current path */
    int    i, k, n;             /* loop variable, number of vertices */
    Vertex v, p, q;             /* corners of a cut-off triangle */
    Edge[] edges;               /* edges of current path */

    if (this.plen <= 0) {       /* if path cuts off the anchor, */
      edges    = new Edge[1];   /* the path contains only one edge */
      edges[0] = new Edge(this.frst, this.last); return edges;
    }                           /* create and return the only edge */
    if (!this.onep)             /* if this is a two-path subproblem */
      edges = new Edge[i = this.plen];
    else {                      /* if this is a one-path subproblem */
      v = this.path[0];         /* get the triangle tip and */
      p = this.key[0];          /* the vertices of the triangle base */
      q = this.key[i = k = 1];    /* get the base of the triangle */
      if (q == null) { q = this.key[i = this.cnt-1]; k = -1; }
      if (v.id < this.verts.length) {  /* if pure triangle split */
        n = this.verts.length;  /* get modulo for id computations */
        if (v.id == (p.id +n -k) % n) {
          edges = new Edge[1];  /* if there is only one subproblem */
          edges[0] = new Edge(q, v); return edges;
        }                       /* create and return triangle side */
        if ((v == this.key[i+k])/* if tip is vertex after base */
        ||  ((q.id < n)         /* or there is only one key edge */
        &&   (v.id == (q.id +n +k) % n))) {
          edges = new Edge[1];  /* if there is only one subproblem */
          edges[0] = new Edge(p, v); return edges;
        }                       /* create and return triangle side */
      }                         /* (otherwise path processing works) */
      edges = new Edge[i = this.plen +1];
      edges[--i] = new Edge(q, v);
    }                           /* add triangle side not on the path */
    while (--i > 0)             /* create edges of the splitting path */
      edges[i] = new Edge(this.path[i-1], this.path[i]);
    edges[0] = new Edge(this.key[0], this.path[0]);
    return edges;               /* add the first path edge and */
  }  /* edges() */              /* return the created edges */

}  /* class Combined1 */
