/*----------------------------------------------------------------------
  File    : Vertex.java
  Contents: management of vertices (perimeter and holes) of a pointgon
  Author  : Christian Borgelt
  History : 2005.02.18 file created
            2005.02.25 functions isLeftOf, isRightOf, isects added
            2005.03.02 function sideOf added
            2005.03.04 function horzLineCuts added
            2005.03.14 function isPartOf added
            2005.03.19 distance to given point (coordinates) added
            2005.06.20 functions isectsX added
            2005.06.23 function isInside added
            2005.12.15 function isInsideX added
            2006.01.21 functions isectsX extended (full edge test)
            2006.02.03 return codes of sideOf inverted
            2006.02.08 epsilon added to function isPartOf
            2013.04.22 adapted to type argument of Comparable
----------------------------------------------------------------------*/
package pointgon;

import java.io.Serializable;

/*--------------------------------------------------------------------*/
/** Class for vertices (perimeter and holes) of a pointgon.
 *  @since  2005.02.18 (Christian Borgelt) */
/*--------------------------------------------------------------------*/
public class Vertex implements Comparable<Vertex>, Serializable {

  private static final long serialVersionUID = 0x00010000L;

  /** vertex identifier */
  public int    id;
  /** x-coordinate of vertex */
  public double x;
  /** y-coordinate of vertex */
  public double y;

  /*------------------------------------------------------------------*/
  /** Create a vertex.
   *  @param  x    the x-coordinate of the vertex
   *  @param  y    the y-coordinate of the vertex
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Vertex (double x, double y)
  { this.x = x; this.y = y; }

  /*------------------------------------------------------------------*/
  /** Compute the distance to a given point.
   *  @param  x    the x-coordinate of the point
   *  @param  y    the y-coordinate of the point
   *  @return the distance to the given point
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double distance (double x, double y)
  {                             /* --- distance to given point */
    x -= this.x; y -= this.y;   /* compute the coordinate diffs. and */
    return Math.sqrt(x*x +y*y); /* from them the Euclidean distance */
  }  /* distance() */

  /*------------------------------------------------------------------*/
  /** Compute the distance to a given vertex.
   *  @param  v    the vertex to compute the distance to
   *  @return the distance to the given vertex
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double distance (Vertex v)
  { return this.distance(v.x, v.y); }

  /*------------------------------------------------------------------*/
  /** Check the position of the vertex relative to an edge.
   *  @param  src   the first  vertex of the edge
   *  @param  dst   the second vertex of the edge
   *  @return -1, if the vertex is in the left, +1, if the vertex is on
   *          the right, and 0 if the vertex is on the edge
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public int sideOf (Vertex src, Vertex dst)
  {                             /* --- check position rel. to edge */
    double a = (dst.x -src.x) *(this.y -src.y);
    double b = (dst.y -src.y) *(this.x -src.x);
    if (a > b) return -1;       /* -1: vertex is to the left */
    if (a < b) return +1;       /* +1: vertex is to the right */
    return 0;                   /*  0: vertex is on the edge */
  }  /* sideOf() */             /*     (i.e. on straight line) */

  /*------------------------------------------------------------------*/
  /** Check the position of the vertex relative to an edge.
   *  @param  src   the first  vertex of the edge
   *  @param  dst   the second vertex of the edge
   *  @return whether the vertex os to the left of the edge
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean isLeftOf (Vertex src, Vertex dst)
  {                             /* --- check position rel. to edge */
    return (dst.x -src.x) *(this.y -src.y)
         > (dst.y -src.y) *(this.x -src.x);
  }  /* isLeft() */

  /*------------------------------------------------------------------*/
  /** Check the position of the vertex relative to an edge.
   *  @param  edge  the edge to check
   *  @return whether the vertex os to the left of the edge
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean isLeftOf (Edge edge)
  { return this.isLeftOf(edge.src, edge.dst); }

  /*------------------------------------------------------------------*/
  /** Check the position of the vertex relative to an edge.
   *  @param  src   the first  vertex of the edge
   *  @param  dst   the second vertex of the edge
   *  @return whether the vertex is to the right of the edge
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean isRightOf (Vertex src, Vertex dst)
  {                             /* --- check position rel. to edge */
    return (dst.x -src.x) *(this.y -src.y)
         < (dst.y -src.y) *(this.x -src.x);
  }  /* isRight() */

  /*------------------------------------------------------------------*/
  /** Check the position of the vertex relative to an edge.
   *  @param  edge  the edge to check
   *  @return whether the vertex is to the right of the edge
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean isRightOf (Edge edge)
  { return this.isRightOf(edge.src, edge.dst); }

  /*------------------------------------------------------------------*/
  /** Check the position of the vertex relative to an edge.
   *  @param  src   the first  vertex of the edge
   *  @param  dst   the second vertex of the edge
   *  @return whether the vertex is on the edge
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean isPartOf (Vertex src, Vertex dst)
  {                             /* --- check for position on edge */
    if (src.x < dst.x) {        /* if left to right edge */
      if ((this.x < src.x) || (this.x > dst.x)) return false; }
    else {                      /* if right to left edge */
      if ((this.x < dst.x) || (this.x > src.x)) return false; }
    if (src.y < dst.y) {        /* if upward edge */
      if ((this.y < src.y) || (this.y > dst.y)) return false; }
    else {                      /* if downward edge */
      if ((this.y < dst.y) || (this.y > src.y)) return false; }
    return Math.abs((dst.x -src.x) *(this.y -src.y)
                  - (dst.y -src.y) *(this.x -src.x)) < 1e-15;
  }  /* isPartOf() */

  /*------------------------------------------------------------------*/
  /** Check the position of the vertex relative to an edge.
   *  @param  edge  the edge to check
   *  @return whether the vertex is on the edge
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean isPartOf (Edge edge)
  { return this.isPartOf(edge.src, edge.dst); }

  /*------------------------------------------------------------------*/
  /** Check whether the vertex is inside a triangle.
   *  @param  a   the first  vertex of the triangle
   *  @param  b   the second vertex of the triangle
   *  @param  c   the third  vertex of the triangle
   *  @return whether the vertex is inside the triangle
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean isInside (Vertex a, Vertex b, Vertex c)
  {                             /* -- check whether inside triangle */
    return !this.isRightOf(a, b)
        && !this.isRightOf(b, c)
        && !this.isRightOf(c, a);
  }  /* isInside() */

  /*------------------------------------------------------------------*/
  /** Check whether the vertex is inside a triangle.
   *  <p>(check both orders of the triangle corners)
   *  @param  a   the first  vertex of the triangle
   *  @param  b   the second vertex of the triangle
   *  @param  c   the third  vertex of the triangle
   *  @return whether the vertex is inside the triangle
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean isInsideX (Vertex a, Vertex b, Vertex c)
  {                             /* -- check whether inside triangle */
    if (c.isLeftOf(a, b))       /* counterclockwise order */
      return !this.isRightOf(a, b)
          && !this.isRightOf(b, c)
          && !this.isRightOf(c, a);
    else                        /* clockwise order */
      return !this.isRightOf(a, c)
          && !this.isRightOf(c, b)
          && !this.isRightOf(b, a);
  }  /* isInsideX() */

  /*------------------------------------------------------------------*/
  /** Check whether a horizontal line through the vertex cuts an edge.
   *  @param  src   the first  vertex of the edge
   *  @param  dst   the second vertex of the edge
   *  @return whether the line cuts the edge
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public int horzLineCuts (Vertex src, Vertex dst)
  {                             /* --- check for line intersection */
    double a, b;                /* buffers for computation */
    Vertex h;                   /* exchange buffer */

    if (src.y == dst.y) {       /* if edge (src,dst) is horizontal */
      if (this.y != src.y)      /* if the edge is out of range, */
        return -2;              /* return 'line does not cut' */
      if (dst.x > src.x) {      /* if edge (src,dst) points to left, */
        h = src; src = dst; dst = h; }    /* exchange the end points */
      return ((this.x >= src.x) && (this.x <= dst.x)) ? 0 : -2;
    }                           /* return whether vertex is on edge */
    if (src.y > dst.y) {        /* if edge (src,dst) leads downward, */
      h = src; src = dst; dst = h; }      /* exchange the end points */
    if ((this.y < src.y) || (this.y >= dst.y))
      return -2;                /* if edge is out of range, abort */
    if (this.y == src.y) {      /* if line cuts lower perimeter joint */
      if     (this.x > src.x) return -1;
      return (this.x < src.x) ? 1 : 0;
    }                           /* return relative position of vertex */
    a = (this.x -src.x) * (dst.y -src.y);
    b = (this.y -src.y) * (dst.x -src.x);
    if (a > b) return -1;     /* check where a horizontal line */
    return (a < b) ? 1 : 0;   /* cuts the edge (vertex on line: 0, */
  }  /* horzLineCuts() */     /* vertex left: -1, vertex right: 1) */

  /* The above function does no general check for an intersection  */
  /* with a horizontal line, but one that is geared to the task of */
  /* checking whether a point is inside a closed path of edges.    */

  /*------------------------------------------------------------------*/
  /** Check whether a line through the vertex cuts an edge.
   *  @param  dx    the x-coordinate of the direction vector of the line
   *  @param  dy    the y-coordinate of the direction vector of the line
   *  @param  src   the first  vertex of the edge
   *  @param  dst   the second vertex of the edge
   *  @return whether the line cuts the edge
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean isects (double dx, double dy, Vertex src, Vertex dst)
  {                             /* --- check for an intersection */
    double a = dx *(src.y -this.y) -dy *(src.x -this.x);
    double b = dx *(dst.y -this.y) -dy *(dst.x -this.x);
    if (a < 0) return (b >= 0); /* check whether a straight line */
    if (a > 0) return (b <= 0); /* through this vertex with direction */
    return true;                /* (dx, dy) cuts the edge (src, dst) */
  }  /* isects() */

  /*------------------------------------------------------------------*/
  /** Check whether a line through the vertex cuts an edge.
   *  @param  to    the vertex the line leads to (from this vertex)
   *  @param  src   the first  vertex of the edge
   *  @param  dst   the second vertex of the edge
   *  @return whether the line cuts the edge
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean isects (Vertex to, Vertex src, Vertex dst)
  { return this.isects(to.x -this.x, to.y -this.y, src, dst); }

  /*------------------------------------------------------------------*/
  /** Check whether a line through the vertex cuts an edge.
   *  @param  to    the vertex the line leads to (from this vertex)
   *  @param  edge  the edge to check
   *  @return whether the line cuts the edge
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean isects (Vertex to, Edge edge)
  { return this.isects(to.x-this.x, to.y-this.y, edge.src, edge.dst); }

  /*------------------------------------------------------------------*/
  /** Check whether a line through the vertex cuts an edge.
   *  @param  to    the vertex the line leads to (from this vertex)
   *  @param  src   the first  vertex of the edge
   *  @param  dst   the second vertex of the edge
   *  @return whether the line cuts the edge
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean isectsX (Vertex to, Vertex src, Vertex dst)
  {                             /* --- check for an intersection */
    double dx = dst.x -src.x;   /* get the direction vector */
    double dy = dst.y -src.y;   /* from source to destination */
    double a  = dx *(this.y -src.y) -dy *(this.x -src.x);
    double b  = dx *(to.y   -src.y) -dy *(to.x   -src.x);
    if (((a < 0) && (b < 0))    /* check whether a straight line */
    ||  ((a > 0) && (b > 0)))   /* through the source with direction */
      return false;             /* (dx, dy) cuts the edge (this, to) */
    dx = to.x -this.x;          /* get the direction vector */
    dy = to.y -this.y;          /* from this vertex to vertex "to" */
    a  = dx *(src.y -this.y) -dy *(src.x -this.x);
    b  = dx *(dst.y -this.y) -dy *(dst.x -this.x);
    if (a < 0) return (b >= 0); /* check whether a straight line */
    if (a > 0) return (b <= 0); /* through this vertex with direction */
    if (b != 0) return true;    /* (dx, dy) cuts the edge (src, dst) */
    return ((src.compareTo(this) *src.compareTo(to) <= 0)
    ||      (dst.compareTo(this) *dst.compareTo(to) <= 0));
  }  /* isectsX() */            /* special test: edges on same line */

  /*------------------------------------------------------------------*/
  /** Check whether a line through the vertex cuts an edge.
   *  @param  to    the vertex the line leads to (from this vertex)
   *  @param  edge  the edge to check
   *  @return whether the line cuts the edge
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean isectsX (Vertex to, Edge edge)
  { return this.isectsX(to, edge.src, edge.dst); }

  /*------------------------------------------------------------------*/
  /** Compare this vertex to another.
   *  @param  obj   the other vertex
   *  @return the sign of the coordinate difference
   *          (x-coordinate takes precedence over y-coordinate)
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public int compareTo (Vertex obj)
  {                             /* --- compare two vertices */
    if (this.x < obj.x) return -1;
    if (this.x > obj.x) return +1;
    if (this.y < obj.y) return -1;
    if (this.y > obj.y) return +1;
    return 0;                   /* return sign of coord. difference */
  }  /* compareTo() */

  /*------------------------------------------------------------------*/
  /** Create a string description of the vertex.
   *  @return the created string description
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public String toString ()
  { return "(" +this.x +", " +this.y +")"; }

}  /* class Vertex */
